/*gdb-tutorial.c, Matthieu Moy pour ENSIMAG*/
#include <stdio.h>

void f(void);
void calcul_et_imprime_fact(int n);
int fact(int n);

/*  Ce fichier est un tutorial pas--pas pour utiliser le dboggeur gdb.
  Un dbogueur est un outil prcieux pour comprendre ce qui ne va pas
  dans vos programmes. Les quelques minutes que vous allez passer 
  apprendre  l'utiliser seront largement rentabilises par la suite !

  Commencez par le compiler. Pour pouvoir utiliser gdb, il faut
  utiliser l'option -g de gcc  la compilation :
  
  $ gcc -g gdb-tutorial.c -o gdb-tutorial

  Vous pouvez lancer l'excutable gdb-tutorial dans gdb en ligne de 
  commande comme ceci :

  $ gdb ./gdb-tutorial

  mais a n'est pas trs convivial. On va donc utiliser le logiciel
  ddd, qui est une interface graphique  gdb. Pour cela, on tape
  simplement

  $ ddd ./gdb-tutorial

  Une autre solution, un peu plus complexe mais trs pratique pour les
  utilisateurs d'Emacs : dans une fentre Emacs, faire :

  M-x gdb RET gdb --annotate=3 gdb-tutorial RET

  (c'est  dire Alt-x, puis "gdb" valid par la touche entre, puis
  "gdb --annotate=3 gdb-tutorial", ce qui devrait tre la commande
  propose par dfaut).

  Dans tous les cas, on se retrouve avec une ligne de commande de gdb.

  Commenons par quelque chose de simple : dmarrer le programme. On
  tape "run", ou simplement la lettre "r" dans la ligne de commande
  gdb. */

int main (int argc, char ** argv) {
  unsigned i;
  int j;
  int * ptr = NULL; /* NULL est une adresse invalide, on n'a pas le
                       droit de drfrencer un pointeur NULL */
  printf ("Bonjour\n");
  printf ("Je vais faire un accs mmoire interdit\n");
  /* 
     La ligne suivante va faire un "Segmentation fault".
     
     Le dboggeur va s'arrter sur cette ligne, et si on est dans 
     Emacs ou ddd, le logiciel ouvre le fichier et nous montre
     la ligne du code sur laquelle notre programme a plant.
  */
  *ptr = 0;
  /*
    Bon, le bug tait facile  trouver. Mettez la ligne ci-dessus en
    commentaire pour liminer le problme, recompilez, et lisez la
    suite.
    On va maintenant faire une execution pas--pas. Il faut commencer
    par donner  gdb un endroit du code o on veut
    s'arrter : un point d'arrt.

    Dans la ligne de commande gdb, on fait "break gdb-tutorial.c:80"
    pour s'arrter  la ligne 80 de gdb-tutorial.c.

    Sous ddd, on peut par exemple double-cliquer dans la marge 
    gauche de la ligne sur laquelle on veut s'arrter.

    Sous Emacs, on met le curseur sur la ligne 81, puis on fait
    C-x SPC (c'est  dire Control-x, puis Espace). Ou bien,  la
    souris, cliquer dans la marge pour ajouter un point d'arrt.

    Faites-le pour la ligne suivante, puis relancez le programme :
  */
  printf ("On avance ...\n"); /* mettre le point d'arrt ici */
  /*
    Maintenant, vous devriez avoir votre programme arrt sur la ligne
    qui prcde. Tapez "next", ou simplement "n" pour avancer dans
    l'execution du programme, plusieurs fois. Petite astuce, aprs
    avoir tap "n" une fois, il suffit d'appuyer sur Entre pour
    refaire un "next".
  */
  printf ("On avance encore ...\n");
  printf ("On avance encore un peu ...\n");
  printf ("On avance encore un peu ...\n");
  printf ("On avance une dernire fois ...\n");
  printf ("Et voil !\n");
  /*
    Il y a deux commandes pour executer un programme pas  pas :
    - next, pour excuter compltement l'instruction suivante
    - step, "rentre" dans la fonction si l'instruction suivante est un
    appel de fonction.
    Essayez successivement "step" et "next" sur les lignes suivantes.
   */
  f();
  f();
  f();
  /*
    Bon, l'execution pas  pas, c'est bien, mais c'est pnible quand
    il y a beaucoup de code  excuter. Par exemple, la boucle 
    suivante prendrait trop de temps, on veut la sauter.
    Pour cela, poser un point d'arrt sur la ligne qui suit la boucle, puis 
    faites "continue" (ou juste "c") dans gdb.
  */
  j = 0;
  for (i = 0; i <= 1000; ++i) {
    j = j + i;
  }
  printf ("Fin de la boucle\n"); /* mettre le point d'arrt ici */
  /*
    On peut bien sr visualiser le contenu d'une variable,
    avec la commande "print", ou simplement "p". On peut
    maintenir l'affichage d'une variable avec "display".

    Par exemple, faites maintenant "display i", puis excutez la
    boucle suivante pas  pas. Faites "p j" pour connaitre la valeur
    de j.

    Sous ddd, on peut aussi voir la valeur d'une variable simplement
    en passant sa souris dessus. Faites un clic droit sur une variable
    dans le source pour voir d'autres possibilits qui vous sont
    offertes. 
  */

  for (i = 1; i <= 1000; i = i * 2) {
    j = j + i;
  }
  /*
    On peut aussi afficher des expressions C un peu plus complexes.
    Essayez par exemple
    (gdb) p i+j
    (gdb) p &i
    (gdb) p &j
    (gdb) p dire_bonjour("Monsieur")

    (dans le dernier cas, gdb va effectivement appeler la fonction
    dire_bonjour, dfinie plus bas)

    On peut aussi s'amuser un peu avec les pointeurs. Faites:

    (gdb) p &i
    $2 = (unsigned int *) 0xbf9bfde8
    
    La valeur 0xbf9bfde8, adresse de i, sera a priori diffrente chez
    vous. Affichez maintenant l'expression *(unsigned int *)0xbf9bfde8
    (en remplaant 0xbf9bfde8 par la valeur que vous avez eu ci-dessus).
    Vous devriez avoir 1024, qui est bien la valeur de i.
  */
  printf ("i=%d\n", i);

  /*
    Tout a se complique un peu quand on a des appels de fonctions.
    Entrez dans la fonction fact appele  la ligne suivante avec la
    commande "s", on se retrouve l bas !
   */
  calcul_et_imprime_fact(10);

  /*
    Une dernire chose : il arrive que le programme parte dans une
    boucle infinie, et on voudrait savoir o est cette boucle.

    Rien de plus simple : lancez le programme ("c" pour "continue"),
    puis tapez Control-c pour arrter son execution et reprendre la
    main dans gdb.

    Sous Emacs, il faut faire deux fois C-c, et sous ddd, vous avez un
    bouton rouge "interrupt".

    Utilisez cette technique pour voir laquelle de ces boucles est une
    boucle infinie.
  */
  j = 1;
  while (j > 0) {
    ++j;
  }

  i = 10;
  while (i >= 0) {
    --i;
  }

  i = 0; j = 0;
  while (i == j) {
    ++i; ++j;
  }
  /*    Voil, vous avez termin et vous savez  peu prs tout ce que vous
    devez savoir pour une utilisation courante. Les plus curieux
    pourront lire le manuel de gdb :
    http://sourceware.org/gdb/download/onlinedocs/  */
}

void f(void) {
  printf ("Appel de f()\n");
  printf ("Fin de f()\n");
}

void calcul_et_imprime_fact(int n) {
  /*
    Re-bonjour,
    
     ce stade, n doit tre gal  10. Vrifiez avec un "p n". Entrez
    dans la fonction fact avec un "step".
  */
  int res = fact(n);
  printf ("fact(%d) = %d\n", n, res);  
}

int fact(int n) {
  printf ("Calcul de fact(%d)\n", n);
  /*
    Si c'est la premire fois que vous passez ici, n doit tre 
    gal  10. Vrifiez avec un "p n". Continuez l'excution pas  pas
    avec "step" (et non "next", pour rentrer dans les appels rcursifs
    successifs de fact). Faites par exemple 5 appels puis lisez le
    commentaire suivant.
  */

  /*
    Voil, je suppose que vous en tes au 5me appel rcursif. Donc,
    "n" doit valoir 5. Vrifiez avec "p n".

    Mais dans la pile, il y a toujours les autres appels, avec les
    valeurs de "n" comprises entre 10 et 5. Faites "where" pour voir
    l'tat de la pile.

    Pour naviguer dans la pile, utilisez les commandes "up" et "down".
    Par exemple, faites deux fois "up", puis "p n". Vous voyez la
    valeur de "n" au 3me appel, c'est  dire 7.

    Faites deux fois "down", et vous voil revenu o vous tiez.
    Faites maintenant "finish" pour terminer
    l'appel de fonction courant, jusqu' revenir  la fonction
    "main".
  */
  if (n <= 1) {
    return 1;
  } else {
    return n * fact(n - 1);
  }
}

int dire_bonjour(char * nom) {
  printf("Bonjour, %s\n", nom);
}
